from datetime import datetime, timedelta
from typing import List
import boto3
from roguewave.modeldata.keygeneration import MODELS, \
    _generate_aws_key, _find_between, _get_model_aws_layout
from functools import cache


def available_models() -> List[str]:
    """
    List all available models
    :return: Available model names as list
    """
    return list(MODELS.keys())


# -----------------------------------------------------------------------------


@cache
def list_available_variables(model_name, init_time: datetime = None):
    """
    List available variables from a given model. This function _only_ works
    if the variable names are encoded in the aws key. For e.g. ECMWF this
    currently does _not_ work.

    Basically, we look on AWS for the pattern given in model.key_template and
    find all keys conforming to the template with different values for the
    variable name. May be very slow if the prefix (everything before the
    variable name in the template) does not sufficiently restrict the search
    space.

    Note response is cached so that multiple calls are fast.

    :param model_name: the name of the model
    :param init_time: init_time, an init_time for which results from the model
    are available. Defaults to 2022-06-01; this may not work for all models.

    :return: list of variables for the model.
    """

    init_time = datetime(2022, 6, 1) if init_time is None else init_time
    # get model aws key template description
    model = _get_model_aws_layout(model_name)

    # open s3 resource
    s3 = boto3.resource('s3')
    my_bucket = s3.Bucket(model.bucket)

    if 'variable' not in model.key_template:
        raise ValueError('Listing variables for a particular model only works'
                         'if the variable name is part of the aws keys.')

    # replace "variable" with "####" as a seperator between everything before
    # and after the variable name in the aws key.
    key = _generate_aws_key(
        '####', init_time, forecast_hour=timedelta(hours=0), model=model)

    # Split in prefix and suffix, and remove the bucket name from the prefix
    prefix, suffix = key.split('####')
    prefix = prefix.replace(model.bucket + '/', '')

    # loop over all objects conforming to the prefix
    variables = []
    for obj in my_bucket.objects.filter(Prefix=prefix):
        if suffix in obj.key:
            # If the suffix is part of the key, add the part in between the
            # prefix and the suffix of the aws key as a variable name to the
            # output list.
            variables.append(
                _find_between(obj.key, prefix, suffix)
            )

    return variables



"""
This file is part of pysofar: A client for interfacing with Sofar Oceans Spotter API

Contents: Convinience functions to handle dates.

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""

from datetime import datetime, timezone
import typing
from roguewave import logger

def to_datetime(time: typing.Union[float, int, datetime, str]):
    if time is None:
        return None

    if isinstance(time, datetime):
        return time.astimezone(timezone.utc)
    elif isinstance(time, str):
        if time[-1] == "Z":
            # From isoformat does not parse "Z" as a valid timezone designator
            time = time[:-1] + "+00:00"
        try:
            return datetime.fromisoformat(time)
        except ValueError as e:
            return datetime.strptime(time, "%Y-%m-%dT%H:%M:%S.%f%z")
    else:
        return datetime.fromtimestamp(time, tz=timezone.utc)

def datetime_to_iso_time_string(time: typing.Union[float, int, datetime, str]):
    if time is None:
        return None

    time = to_datetime(time)
    return time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")



# Import
# =============================================================================
import os
import hashlib
import boto3
from typing import List, Tuple, Dict, Union
from multiprocessing.pool import ThreadPool
from tqdm import tqdm
from roguewave import logger
from pathlib import Path
from warnings import warn

# Model private variables
# =============================================================================


TEMPORARY_DIRECTORY = '~/temporary_roguewave_files/filecache/'
CACHE_SIZE_GiB = 1
MAXIMUM_NUMBER_OF_WORKERS = 10
KILOBYTE = 1024
MEGABYTE = 1024 * KILOBYTE
GIGABYTE = 1024 * MEGABYTE

_ACTIVE_FILE_CACHES = {}


# Classes
# =============================================================================


class AWSFileCache():
    """
    Simple file caching class that when given an aws key locally stores the
    file in the cache directory and returns the path to the file. The file
    remains in storage until the cache directory exceeds a prescribed size,
    in which case files with oldest access/modified dates get deleted first
    until everything fits in the cache again. Any time a file is accessed it's
    modified date gets updated so that often used files automaticall remain in
    cache.

    The files are stored locally in the directory specified on class
    initialization, as:

        [path]/CACHE_PREFIX + md5_hash_of_AWS_KEY

    Methods
      * __getitem__(keys) : accept a simgle aws_key or multiple keys and returns
        filepaths to local copies thereof. You would typically use the
            cache[keys] notation instead of the dunder method.
      * purge() clear all contents of the cache (destructive, deletes all local
        files).

    Usage:

        cache = FileCache()
        list_of_local_file_names = cache[ [list of aws_keys ] ]

    # do stuff with file
    ...
    """

    CACHE_FILE_PREFIX = 'cachefile_'

    def __init__(self,
                 path:str=TEMPORARY_DIRECTORY,
                 size_GiB:Union[float,int]=CACHE_SIZE_GiB,
                 do_cache_eviction_on_startup:bool=False,
                 parallel=True):
        """
        Initialize Cache
        :param path: path to store cache. If path does not exist it will be
            created.
        :param size_GiB: Maximum size of the cache in GiB. If cache exceeds
            the size, then files with oldest access/modified dates get deleted
            until everthing fits in the cache again. Fractional values (floats)
            are allowed.
        :param do_cache_eviction_on_startup: whether we ensure the cache size
            conforms to the given size on startup. If set to true, a cache
            directory that exceeds the maximum size will be reduced to max
            size. Set to False by default in which case an error occurs. The
            latter to prevent eroneously evicting files from a cache that was
            previously created on purpose with a larger size that the limit.


        """

        self.path = os.path.expanduser(path)
        self.max_size = int( size_GiB * GIGABYTE )
        self.parallel = parallel

        # create the path if it does not exist
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

        # Some counters to keep track of total cache misses, hits and
        # evictions. No downstream use right now/
        self._cache_misses = 0
        self._cache_hits = 0
        self._cache_evictions = 0

        # initialize the cache.
        self._entries = {}  # the key/value pair cache
        self._initialize_cache(do_cache_eviction_on_startup)

    def _cache_file_name(self, aws_key: str) -> str:
        """
        Return the filename that corresponds to the given aws_key. We construct
        the file name using a simple md5 hash of the aws_key string prefixed
        with a cache file prefix. THe later is introduced to seperate cache
        files in a path from user files (and avoid including/deleting those).

        :param aws_key: valis aws_key of form "bucket/key"
        :return: valid cache file
        """
        return self.CACHE_FILE_PREFIX + _hashname(aws_key)

    def _cache_file_path(self, aws_key: str) -> str:
        """
        Construct the path where the given aws_key is stored locally.

        :param aws_key: valis aws_key of form "bucket/key"
        :return: valid cache file
        """
        return os.path.join(self.path,
                            self.CACHE_FILE_PREFIX + _hashname(aws_key))

    def _get_cache_files(self) -> List[str]:
        """
        Find all files that are currently a member of the cache
        :return:
        """
        _cache_files = []
        for path, dirs, files in os.walk(self.path):
            # Return all files that are "cache" objects. This is a safety if
            # other user files are present, so that these don't accidentally
            # evicted from the cache (aka deleted).
            return [file for file in files if self.CACHE_FILE_PREFIX if file]
        else:
            return []

    def _initialize_cache(self,do_cache_eviction_on_stratup:bool) -> None:
        """
        Initialize the file cache. Look on disk for files in the cache path
        that have the required prefix and load these into the cache. Once
        loaded, we do a check whether or not the cache is full and if we need
        to remove files.
        :param do_cache_eviction_on_stratup: see description under __init__
        :return:
        """
        self._entries = {}
        for file in self._get_cache_files():
            filepath = os.path.join(self.path, file)
            self._entries[file] = filepath

        # See if cache is still < size
        if do_cache_eviction_on_stratup:
            self._cache_eviction()
        else:
            if self._size() > self.max_size:
                raise ValueError( 'The cache currently existing on disk '
                                  'exceeds the maximum cache size of the '
                                  'current cache.\n Either increase the cache '
                                  'size of the object or allow eviction of '
                                  'files on startup.' )

    def _is_in_cache(self, _hash: str) -> bool:
        """
        Check if a _hash is in the cache
        :param _hash:
        :return:
        """
        cache_hit = _hash in self._entries
        if cache_hit:
            self._cache_hits += 1
        else:
            self._cache_misses += 1
        return cache_hit

    def _add_to_cache(self, _hash: str, filepath: str) -> None:
        # add entry to the cache
        self._entries[_hash] = filepath

    def _get_from_cache(self, _hash:str ) -> str:
        """
        Get entry from cache and touch the file to indicate it has been used
        recently (last to be evicted)
        :param _hash: file_hash corresponding to aws key
        :return: file path
        """
        filepath = self._entries[_hash]
        # Touch the file to indicate we recently used it.
        Path(filepath).touch()
        return filepath

    def __len__(self) -> int:
        """
        :return: Number of entries in the cache.
        """
        return len(self._entries)


    def __getitem__(self, aws_keys: Union[List, str]) -> List[str]:

        # make sure input is a list
        if isinstance(aws_keys, str):
            aws_keys = [aws_keys]

        # Create the hashes from the aws keys
        hashes = [self._cache_file_name(aws_key) for aws_key in aws_keys]

        # collect key, full local path and hash for any hashes that are not
        # in the local cache
        cache_misses = [(key, self._cache_file_path(key), _hash) for
                        _hash, key in zip(hashes, aws_keys) if
                        not self._is_in_cache(_hash)]

        # for all aws_keys not in cache
        if cache_misses:
            # download all the files
            succesfully_downloaded = _download_from_aws(cache_misses,
                                            parallel_download=self.parallel)

            # For all downloaded files do
            for success, cache_miss in zip(
                    succesfully_downloaded, cache_misses):
                if success:
                    # If succesfull, add to cache.
                    self._add_to_cache(cache_miss[2], cache_miss[1])
                else:
                    # If not succesful, remove from keys to return
                    # Todo some logging or erroring?
                    hashes.pop(hashes.index(cache_miss[2]))

        # Get the filepaths to return
        filepaths = [self._get_from_cache(_hash) for _hash in hashes]

        size_of_requested_data = _get_total_size_of_files_in_bytes(filepaths)
        if size_of_requested_data > self.max_size:
            warning = f'The requested data does not fit into the cache.' \
                      f'To avoid issues the cache is enlarged to ensure' \
                      f'the current set of files fits in the cache. \n' \
                      f'old size: {self.max_size} bytes; ' \
                      f'new size {size_of_requested_data+MEGABYTE}'
            warn(warning)
            logger.warning(warning)
            self.max_size = size_of_requested_data + MEGABYTE

        # See if we need to do any cache eviction because the cache has become
        # to big.
        self._cache_eviction()
        return filepaths

    def _cache_eviction(self)->bool:
        """
        Simple cache eviction policy. If the cache exceeds the maximum size
        remove data from the cache based on whichever file was interacted with
        the longest time ago. Evict files until we are below the acceptable
        cache size.

        :return: True if eviction occured, False otherwise.
        """

        # check if we exceed the size, if not return
        if not self._size() > self.max_size:
            return False


        # Get access/modified times for all the files in cache
        modified = []
        for path, dirs, files in os.walk(self.path):
            for _hash in files:
                fp = os.path.join(path, _hash)

                # From my brief reading, access time is not always reliable,
                # hence I use whatever the latest time set is for modified or
                # access time as an indicator of when we last interacted with
                # the file.
                access_time = os.path.getatime(fp)
                modified_time = os.path.getmtime(fp)

                time_to_check = access_time if access_time > modified_time else modified_time
                modified.append((time_to_check, fp, _hash))

        # Sort files in reversed chronological order.
        files_in_cache = [(x[1], x[2]) for x in
                          sorted(modified, key=lambda x: x[0], reverse=True)]

        # Delete files one by one as long as the cache_size exceeds the max
        # size.
        while (_size := self._size()) > self.max_size:
            self._cache_evictions += 1
            logger.debug(
                f'Cache exceeds limits: {_size} bytes, max size: '
                f'{self.max_size} bytes')

            # Get the hash and path of the oldest file
            file_to_delete, _hash = files_in_cache.pop()

            # remove from cache entries.
            self._entries.pop(_hash)
            logger.debug(f' - removing {_hash}')

            # And delete file.
            os.remove(file_to_delete)

        return True

    def _size(self)->int:
        """
        Return size on disk of the cache in bytes.
        :return: cache size in bytes.
        """
        size = 0
        for path, dirs, files in os.walk(self.path):
            size = _get_total_size_of_files_in_bytes(files,path)
        return size

    def purge(self)->None:
        """
        Delete all the files in the cache.
        :return: None
        """
        logger.debug(f'Purging cache')
        keys = list(self._entries.keys())
        for key in keys:
            filepath = self._entries.pop(key)
            logger.debug(f' - deleting {filepath}')
            os.remove(filepath)
        logger.debug(f'Purging cache done')


# Main public function.
# =============================================================================


def cached_local_aws_files( aws_keys:List[str], cache_name:str)-> List[str]:
    """

    :param aws_keys:
    :param cache_name:
    :return:
    """

    if cache_name not in _ACTIVE_FILE_CACHES:
        raise ValueError(f'Cache with name {cache_name} does not exist.')

    return _ACTIVE_FILE_CACHES[cache_name][aws_keys]


def create_aws_file_cache(cache_name:str,
                      cache_path:str=TEMPORARY_DIRECTORY,
                      cache_size_GiB:Union[int,float]=CACHE_SIZE_GiB,
                      do_cache_eviction_on_startup:bool=False,
                      download_in_parallel=True
                      )  \
        -> None:
    """
    Create a file cache. Created caches *must* have unique names and
    cache_paths.

    :param cache_name: name for the cache to be created. This name is used
            to retrieve files from the cache.
    :param cache_path: path to store cache. If path does not exist it will be
            created.
    :param cache_size_GiB:  Maximum size of the cache in GiB. If cache exceeds
            the size, then files with oldest access/modified dates get deleted
            until everthing fits in the cache again. Fractional values (floats)
            are allowed.
    :param do_cache_eviction_on_startup: do_cache_eviction_on_startup: whether we ensure the cache size
            conforms to the given size on startup. If set to true, a cache
            directory that exceeds the maximum size will be reduced to max
            size. Set to False by default in which case an error occurs. The
            latter to prevent eroneously evicting files from a cache that was
            previously created on purpose with a larger size that the limit.
    :param download_in_parallel: Download in paralel from aws. Per default 10
            worker threads are created.

    :return:
    """
    cache_path = os.path.abspath(os.path.expanduser(cache_path))

    if cache_name in _ACTIVE_FILE_CACHES:
        raise ValueError(f'Cache with name {cache_name} is already initialized')

    for key,cache in _ACTIVE_FILE_CACHES.items():
        if cache.path == cache_path:
            raise ValueError(f'Error when creating cache with name: '
                             f'"{cache_name}". \n A cache named: "{key}" '
                             f'already uses the path {cache_path} '
                             f'for caching.\n '
                             f'Multiple caches cannot share the same path.')


    _ACTIVE_FILE_CACHES[cache_name] = AWSFileCache(
            cache_path,
            size_GiB=cache_size_GiB,
            do_cache_eviction_on_startup=do_cache_eviction_on_startup,
            parallel=download_in_parallel
    )
    return


def delete_aws_file_cache(cache_name):
    """
    Delete all files associated with a cache and remove cache from available
    caches. To note: all files are deleted, but the folder itself is not.

    :param cache_name: Name of the cache to be deleted
    :return:
    """
    if cache_name not in _ACTIVE_FILE_CACHES:
        raise ValueError(f'Cache with name {cache_name} does not exist')

    cache = _ACTIVE_FILE_CACHES.pop(cache_name)
    cache.purge()


# Private module helper functions
# =============================================================================


def _hashname(string: str) -> str:
    """
    Returns a md5 hash of a given string.
    :param string: input string
    :return: hexdigest of md5 hash.
    """
    return hashlib.md5(string.encode(), usedforsecurity=False).hexdigest()


def _download_from_aws(key_and_filenames: List[Tuple],
                       parallel_download=False)->List[bool]:
    """
    Wrapper function to download multiple aws_keys from aws.

    :param key_and_filenames: List containing (aws_key,filename)
    :param parallel_download: If true, downloading is performed in parallel.
    :return: List of boolean indicating if the download was a success.
    """
    s3 = boto3.client('s3')
    args = [ (s3,*key_and_filename) for key_and_filename in key_and_filenames ]
    if parallel_download:
        with ThreadPool(processes=MAXIMUM_NUMBER_OF_WORKERS) as pool:
            output = list(
                tqdm(pool.imap(_download_file_from_aws, args),
                     total=len(args)))
    else:
        output = list(
            tqdm(map(_download_file_from_aws, args),
                 total=len(args)))
    return output


def _download_file_from_aws(args):
    """
    Worker function to download files from AWS
    :param args: Tuple/List with as first entry the aws key to download and
        as second entry the filepath to download to
    :return:
    """
    s3 = args[0]
    aws_key = args[1]
    filepath = args[2]
    bucket, key = aws_key.split('/', maxsplit=1)

    s3.download_file(bucket, key, filepath)
    return True

def _get_total_size_of_files_in_bytes(filenames:List[str], path = None) -> int:
    """
    Simple function to calculate the size of a list of files on disk.
    :param filenames: list of filenames or filepaths
    :param path: if filenames are provided, this lists the path, otherwise set
        to None

    :return: Total size in bytes
    """
    size = 0
    for filename in filenames:
        if path is None:
            filepath = filename
        else:
            filepath = os.path.join(path, filename)
        size += os.path.getsize(filepath)
    return size


def _get_cache( path, **kwargs )->AWSFileCache:
    if path in _ACTIVE_FILE_CACHES:
        return _ACTIVE_FILE_CACHES[path]
    else:
        _ACTIVE_FILE_CACHES[path] = AWSFileCache(
            path, **kwargs
        )


from .keygeneration import generate_lead_keys_and_valid_times, \
    generate_forecast_keys_and_valid_times

from roguewave.modeldata.modelinformation import  \
    list_available_variables, available_models

from .pointdata import extract_along_spotter_tracks



import numpy
from datetime import datetime
from numpy import interp
from typing import List,Dict,Union,Tuple

def interpolation_weights(latitude,
                          longitude,
                          model_latitudes,
                          model_longitudes):

    points = enclosing_points(latitude, longitude, model_latitudes, model_longitudes)

    factors = []

    latitude_delta = model_latitudes[points[2][0]] - model_latitudes[points[0][0]]
    longitude_delta = angle_difference(model_longitudes[points[1][1]],model_longitudes[
        points[0][1]])

    #clip latitude
    if ( latitude > 90):
        latitude = 90
    if (latitude < -90):
        latitude = -90

    for point in points:
        longitude_diff = (model_longitudes[point[1]] - longitude + 180) % 360 - 180
        factor = (
            1 - (abs(model_latitudes[point[0]] - latitude)) / latitude_delta
        ) * (1 - abs(longitude_diff) / longitude_delta)

        factors.append({"weight": factor, "lat_lon_indices": point})
    return factors


def enclosing_points(latitude, longitude,
                     model_latitudes,
                     model_longitudes):

    # use proper search algo for sorted arrays
    ii = numpy.searchsorted(  model_latitudes, latitude,side='right' )
    if ii==0:
        ilat = (0,1)
    elif ii == len(model_latitudes):
        ilat = (ii - 2, ii-1)
    else:
        ilat = (ii-1,ii)

    if model_longitudes[0] < 0:
        # longitudes are assumed in [-180,180)
        longitude = (longitude + 180)% 360 - 180
    else:
        # longitudes are assumed in [0, 360)
        longitude = longitude % 360

    if longitude > model_longitudes[-1]:
        longitude = longitude-360

    ii = numpy.searchsorted( model_longitudes, longitude, side='right' )
    if ii == 0:
        ilon = (len(model_longitudes)-1,ii)
    elif ii == len(model_longitudes):
        ilon = (len(model_longitudes)-1,ii)
    else:
        ilon = (ii - 1,ii)

    points = []
    for ii in ilat:
        for jj in ilon:
            points.append((ii, jj))

    return points

def interp_latitude_longitude(obs_time: list[datetime], lat_or_lon,
                              time: List[datetime]):
    # Interpolate latitude or longitude in time making sure we interpolate
    # correctly across the dateline

    # convert times to epochs for interpolation
    obs_time = [x.timestamp() for x in obs_time]
    time = [x.timestamp() for x in time]

    # Make sure we are in [0,360]
    lat_or_lon = numpy.array(lat_or_lon % 360)

    # unwrap to continuous vector
    lat_or_lon = numpy.array(
        numpy.unwrap(lat_or_lon, discont=360, period=360))

    # interpolate and return to [-180,180) range
    return (interp(time, obs_time, lat_or_lon, right=numpy.nan,
                  left=numpy.nan) + 180) % 360 - 180


def angle_difference( angle_to_subtract_from, angle_to_subtract, period=360 ):
    return (angle_to_subtract_from-angle_to_subtract + period/2)%period - period/2


# Import
# =============================================================================
from datetime import datetime, timedelta
from .timebase import ModelTimeConfiguration, timebase_forecast, \
    timebase_lead, timebase_evaluation
import os
from typing import List, Tuple
import json

# Model private variables
# =============================================================================

# Load the model configuration file.
try:
    with open(os.path.expanduser('~/model_configuration.json'), 'rb') as file:
        MODELS = json.load(file)
except FileNotFoundError as e:
    MODELS = {}


# Classes
# =============================================================================
class _ModelAwsKeyLayout:
    """
    Class containing the key template and other necessary information to
    reconstruct keys on AWS.
    """

    def __init__(self,
                 name: str,
                 bucket: str,
                 key_template: str,
                 filetype: str = 'netcdf',
                 model_time_configuration: dict = None):

        """
        :param name: Model name
        :param bucket: bucket
        :param key_template: template of the AWS key
        :param filetype: filetype of the data
        :param model_time_configuration: Description on the timebase of the
        model (how often is the mode run, at what interval is output etc.).
        """
        self.name = name
        self.bucket = bucket
        self.key_template = key_template
        self.filetype = filetype

        if model_time_configuration is None:
            self.model_time_configuration = ModelTimeConfiguration()
        else:
            self.model_time_configuration = ModelTimeConfiguration(
                **model_time_configuration
            )


# Main Public Functions
# =============================================================================
def generate_forecast_keys_and_valid_times(
        variable, init_time: datetime,
        duration: timedelta,
        model_name: str) -> Tuple[List[str], List[datetime]]:
    """
    Get the AWS keys and valid times associated with a forecast for a given
    model.

    :param variable: name of the variable of interest
    :param init_time: init time of the forecast of interest
    :param duration: maximum lead time of interest
    :param model_name: model name
    :return: two results, where the first is a list of the aws_keys and the
    second a list of the valid times that apply to the keys.
    """
    aws_key_layout = _get_model_aws_layout(model_name)
    time_vector = timebase_forecast(
        init_time=init_time,
        duration=duration,
        time_configuration=aws_key_layout.model_time_configuration
    )

    return _generate_keys(time_vector, variable, aws_key_layout)


# ------------------------------------------------------------------------------


def generate_analysis_keys_and_valid_times(
        variable, start_time: datetime, end_time: datetime,
        model_name: str,
) -> Tuple[List[str], List[datetime]]:
    """
    Get the AWS keys and valid times associated with the analysis results for
    a given model.

    :param variable: name of the variable of interest
    :param start_time: Start of time interval of interest
    :param end_time: End of time interval of interest
    :param model_name: model name
    :return: two results, where the first is a list of the aws_keys and the
    second a list of the valid times that apply to the keys.
    """
    aws_key_layout = _get_model_aws_layout(model_name)
    time_vector = timebase_lead(
        start_time,
        end_time,
        timedelta(hours=0),
        time_configuration=aws_key_layout.model_time_configuration,
        exact=False
    )

    return _generate_keys(time_vector, variable, aws_key_layout)


# -----------------------------------------------------------------------------


def generate_lead_keys_and_valid_times(
        variable, start_time: datetime, end_time: datetime,
        lead_time: timedelta, model_name: str,
        exact=False) -> Tuple[List[str], List[datetime]]:
    """
    Get the AWS keys and valid times associated with a given lead time for a
    given model.

    :param variable: name of the variable of interest
    :param start_time: Start of time interval of interest
    :param end_time: End of time interval of interest
    :param lead_time: Lead time of interest
    :param model_name: model name
    :param exact: If true, generate keys exactly at given lead time. Otherwise
    generate keys as close as possible.
    :return: two results, where the first is a list of the aws_keys and the
    second a list of the valid times that apply to the keys.
    """
    aws_key_layout = _get_model_aws_layout(model_name)
    time_vector = timebase_lead(
        start_time=start_time,
        end_time=end_time,
        lead_time=lead_time,
        time_configuration=aws_key_layout.model_time_configuration,
        exact=exact
    )

    return _generate_keys(time_vector, variable, aws_key_layout)


# -----------------------------------------------------------------------------


def generate_evaluation_time_keys_and_valid_times(
        variable,
        evaluation_time, model_name: str,
        maximum_lead_time: timedelta = None
) -> Tuple[List[str], List[datetime]]:
    """

    :param variable: name of the variable of interest
    :param evaluation_time: evaluation time of interest
    :param model_name: model name
    :param maximum_lead_time: maximum lead time of interest
    :return:
    """
    aws_key_layout = _get_model_aws_layout(model_name)
    if maximum_lead_time is None:
        maximum_lead_time = aws_key_layout.model_time_configuration.duration

    time_vector = timebase_evaluation(
        evaluation_time=evaluation_time,
        maximum_lead_time=maximum_lead_time,
        time_configuration=aws_key_layout.model_time_configuration
    )

    return _generate_keys(time_vector, variable, aws_key_layout)


# ------------------------------------------------------------------------------


# Internal module functions
# =============================================================================
def _generate_keys(time_vector, variable, aws_key_layout):
    aws_keys = []
    valid_time = []
    for time in time_vector:
        aws_keys.append(
            _generate_aws_key(
                variable=variable,
                init_time=time[0],
                forecast_hour=time[1],
                model=aws_key_layout)
        )
        valid_time.append(time[0] + time[1])
    return aws_keys, valid_time


def _generate_aws_key(variable, init_time: datetime, forecast_hour: timedelta,
                      model: _ModelAwsKeyLayout) -> str:
    """
    Function that generates a valid AWS key for the given model that represents
    date from the model initialized at the given init_time and with the given
    forecast_hour or lead time.

    :param variable: variable name we want to retrieve (as used in key, if
        applicable, set to None otherwise)
    :param init_time: init time of the model we want to retrieve
    :param forecast_hour: the forecast hour of the model we want to retrieve
    :param model: Model() information on where data is stored.
    :return: A valid aws key for the given model.
    """

    # Helper functions. These are local for now and tied to this function
    # alone. Hence the encapsulation.
    def _replace(_string: str, key: str, value: str):
        """
        Replace a key in string with value
        :param _string: target
        :param key:  string to replace
        :param value:  string to replace with
        :return: string
        """
        return _string.replace("{" + key + "}", value)

    #

    def _time(_string, key, value: datetime):
        """
        Replace {key:fmt} with value.strftime(fmt)
        :param _string: target
        :param key:  either "init_time" or "valid_time"
        :param value: init_time or valid_time as datetime
        :return: string
        """

        # There may be multiple instances that need to be replaced:
        while key in _string:
            # Get the time format string
            _format = _find_between(_string, '{' + key + ':', '}')
            # get the text we want to replace
            to_replace = key + ':' + _format
            # what we want to replace it with (the formatted time)
            replace_with = value.strftime(_format)
            # replace
            _string = _replace(_string, to_replace, replace_with)
        return _string

    #

    def _ecmwf_special(_string, key, _init_time: datetime,
                       _forecast_hour: timedelta):
        """
        Special parsing only valid for ecmwf formatted output.
        :param _string: string
        :param key:  ecmwf_subcycle_string or ecmwf_subcycle_string
        :param _init_time: init time as datetime
        :param _forecast_hour: forecast hour as timedelta
        :return:
        """

        if key == 'ecmwf_subcycle_string':
            # the suffix of ecmwf forecasts is either `001` for all forecast
            # hours exept the zero hour.
            if _forecast_hour == timedelta(hours=0):
                ecmwf_lead_zero_indicator = '011'
            else:
                ecmwf_lead_zero_indicator = '001'
            return _replace(_string, key, ecmwf_lead_zero_indicator)

        elif key == 'ecmwf_subcycle_string':
            # the prefix of ecmwf forecasts is A2P for 10 day forecasts (00Z
            # and # 12Z cycles), or A2Q for the 4 day forecasts (06Z and 18Z
            # cycles)
            if (_init_time.hour == 0) or (_init_time.hour == 12):
                ecmwf_subcycle_string = 'A2P'
            else:
                ecmwf_subcycle_string = 'A2Q'
            return _replace(_string, key, ecmwf_subcycle_string)
        else:
            raise Exception(f'Unknown key {key}')

    # ---------------------

    #
    # Purpose:
    #
    #   The target template key has the form
    #   "t/{init_time:%Y%m%d}/{init_time:%H}/t/{forecasthour}.{variable}.nc"
    #   here everything between {...} defines a token that needs to be replaced
    #   at executing time by a value. E.g. {init_time:%Y%m%d} denotes we want
    #   to replace this with the init_time formated according to %Y%m%d. This
    #   function does the parsing/replacement.
    #
    # Implementation:
    #
    #  We define a set of properties below, where the key denotes the keyword
    #  we want to replace (init_time, forecast_hour etc.), and the value is
    #  a dictionary with a method denoting the function handling the
    #  replacement and a set of arguments in addition to the string and the key
    #  to be replaced.
    valid_time = init_time + forecast_hour

    # Define keywords
    keywords = {
        "bucket": {'method': _replace, 'args': (model.bucket,)},
        "forecasthour": {
            'method': _replace, 'args': (
                "{hours:03d}".format(
                    hours=int(forecast_hour.total_seconds() // 3600)
                ),
            )
        },
        "init_time": {'method': _time, 'args': (init_time,)},
        "valid_time": {'method': _time, 'args': (valid_time,)},
        "variable": {'method': _replace, 'args': (variable,)},
        "ecmwf_subcycle_string": {
            'method': _ecmwf_special, 'args': (init_time, forecast_hour)
        },
        "ecmwf_lead_zero_indicator": {
            'method': _ecmwf_special, 'args': (init_time, forecast_hour)
        }
    }

    # Loop over properties and invoke method to replace keywords.
    string = model.key_template
    for keyword, action in keywords.items():
        if keyword in string:
            string = action['method'](string, keyword, *action['args'])

    return f"{model.bucket}/{string}"


# Helper functions
# =============================================================================
def _find_between(s: str, first: str, last: str) -> str:
    """
     Simple function that returns the string between the "first" and "last"
     token. E.g. if first='{' and last=']' with string 'asdd{THIS]asdsd'
     this returns 'THIS'.
    :param s: string
    :param first: token signifying start
    :param last:  token signifying end
    :return:  string
    """

    # lets avoid regex for now :-)
    try:
        start = s.index(first) + len(first)
        end = s.index(last, start)
        return s[start:end]
    except ValueError:
        return ""


# -----------------------------------------------------------------------------


def _get_model_aws_layout(name) -> _ModelAwsKeyLayout:
    """
    Return the object representing key layout on aws
    :param name:
    :return: Model
    """
    return _ModelAwsKeyLayout(name=name, **MODELS[name])

# -----------------------------------------------------------------------------
